#ifndef COMMON_HPP
#define COMMON_HPP

#include <bitset>

typedef std::tuple<uint64_t, uint64_t, uint64_t, uint64_t> tiiii;
typedef std::tuple<uint64_t, uint64_t> tii;
typedef sdsl::bit_vector::iterator	bit_iterator;

/* A Save version of the SDSL Semi extern PHI LCP construction algorithm, that checks Textboundaries */
void construct_lcp_semi_extern_PHI_save(sdsl::cache_config& config)
{
    using namespace sdsl;
    using namespace std;
    typedef int_vector<>::size_type size_type;
    int_vector_buffer<> sa_buf(cache_file_name(conf::KEY_SA, config));
    size_type n = sa_buf.size();
    if (1==n) {
        int_vector<> lcp(1, 0);
        store_to_cache(lcp, conf::KEY_LCP, config);
        return;
    }
    const uint8_t log_q = 6; // => q=64
    const uint32_t q = 1<<log_q;
    const uint64_t modq = bits::lo_set[log_q];

    // n-1 is the maximum entry in SA
    int_vector<64> plcp((n-1+q)>>log_q);

    for (size_type i=0, sai_1=0; i < n; ++i) {   // we can start at i=0. if SA[i]%q==0
        // we set PHI[(SA[i]=n-1)%q]=0, since T[0]!=T[n-1]
        size_type sai = sa_buf[i];
        if ((sai & modq) == 0) {
//            if ((sai>>log_q) >= plcp.size()) {
//                    std::cerr<<"sai="<<sai<<" log_q="<<log_q<<" sai>>log_q="<<(sai>>log_q)<<" "<<sai_1<<std::endl;
//                    std::cerr<<"n="<<n<<" "<<" plcp.size()="<<plcp.size();
//            }
            plcp[sai>>log_q] = sai_1;
        }
        sai_1 = sai;
    }
    plcp[sa_buf[0] >> log_q] = n;

    int_vector<8> text;
    load_from_cache(text, conf::KEY_TEXT, config);

    for (size_type i=0,j,k,l=0; i < plcp.size(); ++i) {
        j =	i<<log_q;   // j=i*q
        k = plcp[i];
        while ( (j+l) < n && (k+l) < n && text[j+l] == text[k+l])
            ++l;
        plcp[i] = l;
        if (l >= q) {
            l -= q;
        } else {
            l = 0;
        }
    }

    size_type buffer_size = 4000000; // buffer_size is a multiple of 8!
    sa_buf.buffersize(buffer_size);
    int_vector_buffer<> lcp_out_buf(cache_file_name(conf::KEY_LCP, config), std::ios::out, buffer_size, sa_buf.width());	// open buffer for plcp

    for (size_type i=1, sai_1=sa_buf[0],l=0, sai=0,iq=0; i < n; ++i) {
        /*size_type*/ sai = sa_buf[i];
//				std::cerr<<"i="<<i<<" sai="<<sai<<std::endl;
        if ((sai & modq) == 0) { // we have already worked the value out ;)
            lcp_out_buf[i] = l=plcp[sai>>log_q];
        } else {
            /*size_type*/ iq = sai & bits::lo_unset[log_q];
            l  = plcp[sai>>log_q];
            if (l >	(sai-iq))
                l -= (sai-iq);
            else
                l=0;
            while ( (sai+l) < n && (sai_1+l) < n && text[ sai+l ] == text[ sai_1+l ])
                ++l;
            lcp_out_buf[i] = l;
        }
        sai_1 = sai;
    }
    lcp_out_buf.close();
    register_cache_file(conf::KEY_LCP, config);
    return;
}


uint64_t huffman_code_length(const std::vector<uint64_t>& count)
{
	std::priority_queue<uint64_t, std::vector<uint64_t>, std::greater<uint64_t>> pq(begin(count), end(count));
	uint64_t result = 0;
	while(pq.size() > 1)
	{
		uint64_t w = pq.top();
		pq.pop();
		w += pq.top();
		pq.pop();
		pq.push(w);
		result += w;
	}
	return result;
}

std::pair<uint64_t, uint64_t> rice_code_length(const std::vector<int64_t>& data, bool riceonly=true)
{
	uint64_t best_p = 0;
	uint64_t best_len = 1<<31;
	
	for (uint64_t p = 1; p < 30; p++)
	{
		uint64_t len = 0;
		for (auto l: data)
		{
			uint64_t q = abs(l) >> p;
			if (not riceonly)
				len += 2;			// Sign&diffBits
			len += q+1 + p;			// full part (Unär) + rest
		}		
		
		if (len < best_len)
		{
			best_p = p;
			best_len = len;
		}
	}
	
	return {best_p, best_len};
}

std::tuple<uint64_t, uint64_t, uint64_t> lencompress(sdsl::int_vector<>& lengths, uint64_t x)
{
	uint64_t maxlen = 1 << x;
	uint64_t counter = 0;
	size_t bits = 0;
	size_t width = lengths.width();
	
	std::vector<int64_t> data;
	
	for (auto l: lengths)
	{
		if (l < maxlen)
		{
			bits += 1 + x;
			counter++;	
			data.push_back(l);
		}
		else
			bits += 1 + width;
	}
	
	auto rice = rice_code_length(data, true);
	rice.second += data.size() + (lengths.size() - data.size()) * (width + 1);
	
	if (rice.second < bits)
		return std::make_tuple(rice.second, counter, rice.first);
	return std::make_tuple(bits, counter, 0);
}

// Append the rice code of [num] to the vector [code] with the parameter [p]
bool append_ricecode(uint64_t num, uint64_t p, std::vector< bool >& code)
{
	uint64_t q = num >> p;
	
	for (uint64_t i = 0; i < q; i++) code.push_back(1);
	code.push_back(0);
	
	uint64_t b = 1 << p;
	uint64_t r = num & (b-1);
		
	std::bitset<64> bs(r);	
	for (uint64_t i = p; i > 0; i--) 
	{ 
		code.push_back(bs[i-1]);
	}
	
	return true;
}



bool binary_search(bit_iterator& begin, bit_iterator& end)
{
	uint64_t mid = uint64_t(end - begin) / 2;
	auto middle = begin;
	std::advance( middle, mid );
	
	if (*begin == 1 and *(end-1) == 1 and *middle == 1)
	{
		return false;
	}
	
	if (*begin == 0 and *(end-1) == 0)
	{
		if (*middle == 0)
		{
			end--;
			return true;
		}
		// TODO... Ignore for now. (Break in the middle)
		auto tmp = begin;
		begin = end;
		end = tmp;
		return false;
	}
	
	// Find first...
	if (*begin == 1)
	{
		begin = std::upper_bound(begin, end, 1, std::greater<int>());
	}
	
	// and second
	if (*(end-1) == 1)
	{
		end = std::lower_bound(begin, end, 1);
	}
	
	end--;
	return true;
}


uint64_t read_ricecode(uint64_t& idx, uint64_t p, sdsl::int_vector<>& code)
{	
	uint64_t n;	
	uint64_t q = 0;
	while (code[idx++] == 1) q++;
	n = q << p;
	
	
	//uint64_t b = 1 << p;
	uint64_t r = 0;
	for (uint64_t i = 0; i < p; i++)
	{
		r = r << 1;
		r += code[idx++];
	}
	n += r;
	return n;
}

int64_t get_next_mapping_start(std::tuple<uint64_t, uint64_t, uint64_t>& idx, 
							   sdsl::bit_vector& bv, sdsl::int_vector<>& abs, sdsl::int_vector<>& rel,
							   uint64_t prev)
{
	int64_t code = 0;
	if (bv[std::get<0>(idx)++] == 0) 
	{	// Absolut code
		code = abs[std::get<1>(idx)++];
	}
	else
	{
		if (bv[0] == 0) code = rel[std::get<2>(idx)++];
		else code = read_ricecode(std::get<2>(idx), abs[0], rel);
		if (bv[std::get<0>(idx)++] == 1) code *= -1;
		
		code = prev + code;
	}
    return code;
}

int64_t get_next_mapping_len(std::tuple<uint64_t, uint64_t, uint64_t>& idx, 
							 sdsl::bit_vector& bv, sdsl::int_vector<>& longs, sdsl::int_vector<>& shorts)
{
	int64_t code = 0;
	if (bv[std::get<0>(idx)++] == 0) 
	{	// Absolut code
		code = longs[std::get<1>(idx)++];
	}
	else
	{
		if (bv[0] == 0) code = shorts[std::get<2>(idx)++];
		else code = read_ricecode(std::get<2>(idx), longs[0], shorts);
	}
	
    return code;
}


#endif
