#ifndef COMPRESS_BASIC_HPP
#define COMPRESS_BASIC_HPP

#include "construct.hpp"
#include "sdsl/io.hpp"
#include <cstdlib>
#include <map>
#include <fstream>
#include <sstream>
#include "sdsl/enc_vector.hpp"
#include <sdsl/util.hpp>
#include <bitset>
#include "common.hpp"

#define SPECIALCHAR 1
const uint8_t F_bitsize = 5;

void compress_basic(sdsl::cache_config& config, std::string outfile, uint64_t a, uint64_t b=0, std::string comp="bzip")
{	
    using namespace std;
    using namespace sdsl;
    // Build LCP Array and sort by length
	pqli pq = bu_lcp(config, a);
	    
    // Load sa from cache
    int_vector<0> sa;
    load_from_cache(sa, conf::KEY_SA, config);
    // Load sa from cache
    int_vector<0> lcp;
    load_from_cache(lcp, conf::KEY_LCP, config);
            
    // Load sa from cache
    int_vector<8> original;
    load_from_cache(original, conf::KEY_TEXT, config);
    
            
    uint64_t n = sa.size();
                                    
    bit_vector done(n, 0);
    
    vector<tiiii> factors;
    // UID == 0 => Don't replace anything
    // UID == 1 => Repeat
    // UID == 2 => New factor
	uint64_t uid = 3;
	
	uint64_t k = 0;
	
    vector<uint64_t> accepted;       
    vector<uint64_t> candidates;   
	vector<uint64_t> others;
	
	uint64_t interval_size = 0;
    for(; !pq.empty(); ++k)                  
    {
    
		accepted.resize(0);
		candidates.resize(0);
		others.resize(0);
		    
		lcpinterval li = pq.top();
		pq.pop();
    	interval_size += li.rb - li.lb + 1;
						
		uint64_t l = li.l;		
        
        // occ1 = min{SA[i] | lb <= i <= rb}
        uint64_t occ1 = n; bool occ1_candidates = false;
        
        // occ2 = min(candidates \ {occ1})
        uint64_t occ2 = n;
        
        // occ3 = min(accepted \ {occ1})
        uint64_t occ3 = n;
        
        for (uint64_t i = li.lb; i <= li.rb; i++)
        {             
            auto sa_i = sa[i];         
            bool accept = done[sa_i] == 0 && done[sa_i+l-1] == 0; // && done[sa_i+l/2] == 0; 
              
            // Determin minimum  (i.e. occ1 = min{SA[i] | lb <= i <= rb} )
            if (sa_i < occ1)
            {
                if (occ1 < occ2 and occ1_candidates) occ2 = occ1;   // Check if this maybe occ2
                occ1 = sa_i;
                occ1_candidates = accept;
            }
            
            if (accept)
            {
                candidates.push_back(sa_i);
                if (sa_i < occ2 and sa_i > occ1) occ2 = sa_i;
            }
        }
        
               
        uint64_t s = candidates.size();
        
        if (s + !occ1_candidates <= 1) continue;       // Nothing to compress
        
        if (true) // (l >= minlen_real)        // Enough Candidates?
        {   
            sort(candidates.begin(), candidates.end());
                        
            if (!occ1_candidates)
                accepted.push_back(candidates[0]);
                
            for (uint64_t j=0; j < s-1; j++)
            {
                if (candidates[j] + li.l <= candidates[j+1])
                {
                    accepted.push_back(candidates[j+1]);
                    if (candidates[j+1] < occ3 and candidates[j+1] > occ1) occ3 = candidates[j+1];
                }
            }
            uint64_t t = occ2 - occ1;
            if (occ1_candidates and t < l and occ3 != occ2 and (occ3 - occ2) >= l)
            {            
                factors.push_back( make_tuple(occ1+t, 1, t, l) );
                for (uint64_t i = 0, start=occ1+t; i < l; i++)
                    done[start+i] = 1;
            }
            
            s = accepted.size();
                                                   
            if (s <= 0) continue;
            
    		uint64_t minlen_real = uint64_t(a / s) + b;
    		
            if (l >= minlen_real)
            {       
                uint64_t s = accepted.size();
                uint64_t id = 0;
                if (s == 1) id = 2;
                else        id = uid++;     
                uint64_t start = 0;   
                                    
                for (; start < s; start++)
                {                    
                    auto c = accepted[start];      
                    factors.push_back( make_tuple(c, id, occ1, l) );                   
                    for (uint64_t i=0; i < l; i++)
                    {
                        done[c+i] = 1;
                    }
                }
            }   
        }
    }
    sort(factors.begin(), factors.end());
           
    uint64_t uid_remapped = 3;
    vector< tuple<uint64_t, uint64_t> > mapping;
    
    // Helper for reordering uids
    uint8_t uid_width = sdsl::bits::hi(uid+1)+1;
    int_vector<0> helper(uid, uid, uid_width);
    int_vector<64> firstocc(uid, uid);
    
    // F vector
    //uint64_t F_size = factors.size() - uid + 1;
    //int_vector<0> F(F_size, 0, uid_width);
    vector<uint64_t> F_new;
    
    // Counter for F values for huffman
    vector<uint64_t> count(uid, 0);
    
    uint64_t i = 0;
    
    vector<char> S_new;
    
    // Build F vector
    uint64_t factor_idx = 0;
    uint64_t factor_size = factors.size();
	// Load text from cache
    int_vector<8> text;
    load_from_cache(text, conf::KEY_TEXT, config);
    
    int_vector<> counter(256);
    for (auto c : text) 
    {   
        counter[c]++;
    }
    uint64_t min = text.size() + 1;
    uint64_t argmin = 0;
    for (uint64_t c=0; c < 256; c++)
    {
        if (counter[c] < min)
        {   
            argmin = c; 
            min = counter[c];
        }
        
    }
    unsigned char specialchar = (unsigned char)argmin;
    
    S_new.push_back(specialchar);         // First char is specialchar.
        
    for (; factor_idx < factor_size; factor_idx++)
    {        
        auto f = factors[factor_idx];
        uint64_t pos = get<0>(f);
                
        while (i < pos)
        {
            S_new.push_back(text[i]);
            
            if (text[i] == specialchar)
            {
                F_new.push_back(0);
            }
            
            i++;
        }
        
        uint64_t id = get<1>(f);
    
        if (id == 1 or id == 2)     // Special case
        {
            mapping.push_back( make_tuple(get<2>(f), get<3>(f)) );
            
            F_new.push_back(id);     // Special case ( repeat or only one occ)
             
        }
        
        else if (helper[id] == uid)      // First occ of a factor
        {
            mapping.push_back( make_tuple(get<2>(f), get<3>(f)) );
            
            helper[id] = uid_remapped;
            uid_remapped++;
            F_new.push_back(id);
        }
        else
        {
        	F_new.push_back(id);
        }
                    
        // Update S and pos
        S_new.push_back(specialchar);
        i += get<3>(f);
    }
    
    
    while (i < n)
    {
        S_new.push_back(text[i]);
        
        if (text[i] == specialchar)
        {
		    F_new.push_back(0);
        }
        
        i++;
    }    
    	
	/*================*/
    /* Compress files */
    /*================*/
	    
    uint64_t msize = mapping.size();
    int_vector<> mapping_start(msize);
    int_vector<> mapping_length(msize);
    
    for (uint64_t i=0; i < msize; i++){
        mapping_start[i] = get<0>(mapping[i]);
        mapping_length[i] = get<1>(mapping[i]);
    }
    util::bit_compress(mapping_start);
    util::bit_compress(mapping_length);
                
    uint64_t Ssize = S_new.size();
    int_vector<8> S(Ssize);
    for (uint64_t i=0; i < Ssize; i++){
        S[i] = S_new[i];
    }
    

    uint64_t Fsize = F_new.size();
    int_vector<> F(Fsize);
    for (uint64_t i=0; i < Fsize; i++){
        F[i] = F_new[i];
    }
    util::bit_compress(F);
    
    /* =================================== */
    // Try to compress mapping   
    /* =================================== */
	
	/* Get parameters for mapping start compression */
    uint64_t width = mapping_start.width();
    
    uint64_t best_x = 0;
    uint64_t p = 0;
    uint64_t best_method = 0; // 0 == Normal | 1 == Ricecode
    uint64_t best_score = 1L<<31;
    uint64_t num_rel = 0;
    
    for (uint64_t x = 1; x < width-1; x++)
	{
		vector<int64_t> relativs;
		uint64_t maxdif = 1 << x;
    	    	
		uint64_t abs = 0, posrelnegrelend = 0;
		uint64_t posrel = 0, posrelend = 0, posrelboth = 0;
		uint64_t negrel = 0, negrelend = 0, negrelboth = 0;
		
		for (uint64_t i = 1; i < msize; i++)
		{
			uint64_t ms_i = mapping_start[i];
			uint64_t me_i1 = mapping_start[i-1] + mapping_length[i-1];
			uint64_t ms_i1 = mapping_start[i-1];
			bool b_posrel = false, b_posrelend = false, b_negrel = false, b_negrelend = false;
			
			
			if (ms_i >= ms_i1) 		// Positiv from start
				b_posrel = (ms_i - ms_i1) < maxdif;	
			else					// Negativ from start
				b_negrel = (ms_i1 - ms_i) < maxdif;
			
			if (ms_i >= me_i1)		// Positiv from end
				b_posrelend = (ms_i - me_i1) < maxdif;
			else 			 		// Negativ from end
				b_negrelend = (me_i1 - ms_i) < maxdif;
			
				
			
			if (b_posrel && b_posrelend)
				posrelboth++;
			else if (b_posrel && b_negrelend)
				posrelnegrelend++;
			else if (b_posrel)
				posrel++;
			else if (b_posrelend)
				posrelend++;
			else if (b_negrel && b_negrelend)
				negrelboth++;
			else if (b_negrel)
				negrel++;
			else if (b_negrelend)
				negrelend++;
			else if (b_posrel or b_posrelend or b_negrel or b_negrelend)
			{
				cerr << "Unrecognized: " << b_posrel << " " << b_posrelend << " " << b_negrel;
				cerr << " " << b_negrelend << " " << posrelnegrelend << endl;
				
				cerr << ms_i << " " << ms_i1 << " " << me_i1 << " " << maxdif << endl;
			}	
			else
				abs++;
			
			if (b_posrelend)
				relativs.push_back(me_i1 - ms_i);
				
			else if (b_negrelend)
				relativs.push_back(ms_i - me_i1);
				
				
		}
		// Fixed len for relatives: 	(No huffman prefix for now)
		vector<uint64_t> occ(3);
		occ[0] = abs + posrel + negrel;						// Normal case
		occ[1] = posrelend + posrelboth;					// Positiv
		occ[2] = negrelend + negrelboth + posrelnegrelend;	// Negativ
						
		//uint64_t size_normal = huffman_code_length(occ);
		uint64_t size_normal = 0;
		size_normal += (width+1) * occ[0];
		size_normal += (x+2) * (occ[1] + occ[2]);
		
		// Rice 
		//uint64_t size_rice = huffman_code_length(occ);
		uint64_t size_rice = 0;
		auto rice = rice_code_length(relativs);
		
		size_rice += rice.second + (relativs.size() * 2);
		size_rice += (width+1) * occ[0];
		
		if (size_normal < best_score && size_normal < size_rice)
		{
			best_score = size_normal;
			best_x = x;
			best_method = 0;		
			num_rel = occ[1] + occ[2];	
		}
		else if (size_rice < best_score)
		{
			best_score = size_rice;
			p = rice.first;
			best_x = x;
			best_method = 1;	
			num_rel = relativs.size();
		}
	}
	
	/* Compress mapping start with best parameter set */
	sdsl::bit_vector bv(num_rel + msize + 1, 0);
	bv[0] = best_method;		// Write type of short compression (i.e. rice code or bitcompressed)
	uint64_t bv_idx = 1;
		
	vector< bool > ricecode;
	vector< uint32_t > relcodes;
	
	uint64_t abscodesize = msize - num_rel;
	if (best_method == 1) abscodesize++;
	
	sdsl::int_vector<> abscodes(abscodesize);
	uint64_t abs_idx = 0;
	if (best_method == 1) abscodes[abs_idx++] = p;
	
	uint64_t maxdif = 1 << best_x;
		
	// Write first start absolute
	if (abscodesize > 0)
	{
    	abscodes[abs_idx++] = mapping_start[0];
	    bv[bv_idx++] = 0;
    }
	
	for (uint64_t i = 1; i < msize; i++)
	{
		uint64_t ms_i = mapping_start[i];
		uint64_t me_i1 = mapping_start[i-1] + mapping_length[i-1];
		
		bool b_posrelend = false, b_negrelend = false;
		
		uint64_t dif;
			
		if (ms_i >= me_i1)		// Positiv from end
		{
			b_posrelend = (ms_i - me_i1) < maxdif;
			dif = (ms_i - me_i1);
		}
		else 			 		// Negativ from end
		{
			b_negrelend = (me_i1 - ms_i) < maxdif;
			dif = (me_i1 - ms_i);
		}
				
		if (b_posrelend)
		{	// PrefixCode is 10
			bv[bv_idx] = 1;
			bv_idx += 2;				
			if (best_method == 0) relcodes.push_back(dif);
			else append_ricecode(dif, p, ricecode);
		}
		else if (b_negrelend)
		{	// PrefixCode is 11
			bv[bv_idx++] = 1;
			bv[bv_idx++] = 1;			
			if (best_method == 0) relcodes.push_back(dif);
			else append_ricecode(dif, p, ricecode);
		}
		else
		{	// Prefix code is 0
			bv_idx++;
			abscodes[abs_idx++] = mapping_start[i];
		}			
	}
		
	// Store to file
	util::bit_compress(abscodes);
	store_to_file(abscodes, "mapping_start_abs.sdsl");
	store_to_file(bv, "mapping_start_bv.sdsl");			
	if (best_method == 0)
	{	
		uint64_t relcodessize = relcodes.size();
		sdsl::int_vector<> rel(relcodessize);
		for (uint64_t i = 0; i < relcodessize; i++) rel[i] = relcodes[i];
		util::bit_compress(rel);
    	store_to_file(rel, "mapping_start_rel.sdsl");
	}
	else
	{
		uint64_t ricecodesize = ricecode.size();
		sdsl::int_vector<> rice(ricecodesize);
		for (uint64_t i = 0; i < ricecodesize; i++) rice[i] = ricecode[i];
		util::bit_compress(rice);
    	store_to_file(rice, "mapping_start_rel.sdsl");
	}
	   
    /* ======================= */
    /* Compress Mapping Lenght */
    /* ======================= */
	// Find parameter
    auto best = lencompress(mapping_length, 1);
    best_x = 1;
    
    for (int64_t x = 2; x < mapping_length.width() - 1; x++)
    {
    	auto s = lencompress(mapping_length, x);
    	if (get<0>(s) < get<0>(best))
    	{
    		best = s;
    		best_x = x;
    	}
    }
    
	// Write Bitvector and len vectors
    bit_vector bv_len(mapping_length.size() + 1, 0);
    bv_idx = 0;
    uint64_t longsize = mapping_length.size() - get<1>(best);
    if (get<2>(best) != 0)	// Write type of short compression (i.e. rice code or bitcompressed)
    {
		longsize++;
    	bv_len[bv_idx++] = 1;
    }	
    else
    {
        bv_len[bv_idx++] = 0;
    }
    int_vector<> long_len(longsize);
    uint64_t long_idx = 0;
    if (get<2>(best) != 0)
    	long_len[long_idx++] = get<2>(best);
    	
    int_vector<> short_len(get<1>(best));
    vector<bool> rice_len;
    uint64_t short_idx = 0;    
        
    uint64_t maxlen = 1L << best_x;
    
    for (auto l: mapping_length)
    {
		if (l < maxlen)
		{
			bv_len[bv_idx++] = 1;
			if (get<2>(best) == 0)
				short_len[short_idx++] = l;
			else
				append_ricecode(l, get<2>(best), rice_len);
		}
		else
		{
			bv_len[bv_idx++] = 0;
			long_len[long_idx++] = l;
		}
    }    
    
    util::bit_compress(long_len);
    if (get<2>(best) == 0)
    {
	    util::bit_compress(short_len);
		store_to_file(short_len, "mapping_length_short.sdsl");
    }
    else
    {
    	uint64_t ricecodesize = rice_len.size();
		sdsl::int_vector<> rice(ricecodesize);
		for (uint64_t i = 0; i < ricecodesize; i++) rice[i] = rice_len[i];
		util::bit_compress(rice);
    	store_to_file(rice, "mapping_length_short.sdsl");
    }
    	
	store_to_file(long_len, "mapping_length_long.sdsl");
	store_to_file(bv_len, "mapping_length_bv.sdsl"); 
    /* =================================== */
    
    /* =================================== */
    // Store in files    
    /* =================================== */
    store_to_file(S, "S.sdsl");
    store_to_file(F, "F.sdsl");    
        
    auto r = system("tar -cf mapping.tar mapping_length_long.sdsl mapping_length_short.sdsl mapping_length_bv.sdsl "
	       "mapping_start_bv.sdsl mapping_start_abs.sdsl mapping_start_rel.sdsl && "
	       "rm mapping_length_long.sdsl mapping_length_short.sdsl mapping_length_bv.sdsl "
	       "mapping_start_bv.sdsl mapping_start_abs.sdsl mapping_start_rel.sdsl");
    if (r != 0) cerr << "Error while tar mapping" << endl;
    
    std::string cleanup;

    stringstream s;
    s << "tar -cf " << outfile;
    if (!std::strncmp(comp.c_str(), "ppmd", 4))
    {
        r = system("ppmd e -s -d -fmapping.tar.pmd mapping.tar");
        if (r != 0) cerr << "Error while ppmd mapping" << endl;
        r = system("ppmd e -s -d -fF.sdsl.pmd F.sdsl");
        if (r != 0) cerr << "Error while ppmd F" << endl;
        r = system("ppmd e -s -d -fS.sdsl.pmd S.sdsl");
        if (r != 0) cerr << "Error while ppmd S" << endl;
        
        s << " mapping.tar.pmd F.sdsl.pmd S.sdsl.pmd";
        cleanup = "rm -f mapping.tar.pmd F.sdsl.pmd S.sdsl.pmd";
    }
    else if (!std::strncmp(comp.c_str(), "zpaq", 4))
    {
        r = system("~/zpaq/zpaq a mapping.tar.zpaq mapping.tar -m5 -t1; rm mapping.tar");
        if (r != 0) cerr << "Error while zpaq mapping" << endl;
        r = system("~/zpaq/zpaq a F.sdsl.zpaq F.sdsl -m5 -t1; rm F.sdsl");
        if (r != 0) cerr << "Error while zpaq F" << endl;
        r = system("~/zpaq/zpaq a S.sdsl.zpaq S.sdsl -m5 -t1; rm S.sdsl");
        if (r != 0) cerr << "Error while zpaq S" << endl;
        s << " mapping.tar.zpaq F.sdsl.zpaq S.sdsl.zpaq";
        cleanup = "rm -f mapping.tar.zpaq F.sdsl.zpaq S.sdsl.zpaq";
    }
    else
    {
        r = system("xz -9 mapping.tar");
        if (r != 0) cerr << "Error while xz mapping" << endl;
        r = system("xz -9 -f F.sdsl");
        if (r != 0) cerr << "Error while xz F" << endl;
        r = system("xz -9 -f S.sdsl");
        if (r != 0) cerr << "Error while xz S" << endl;
        
        s << " mapping.tar.xz F.sdsl.xz S.sdsl.xz";
        cleanup = "rm -f mapping.tar.xz F.sdsl.xz S.sdsl.xz";
    }   
     
    r = system(s.str().c_str());
	if (r != 0)
	    cerr << "Error while tar all" << endl;
	    
    r = system(cleanup.c_str());
    
}

#endif
