#ifndef DECOMPRESS_ADVANCED_HPP
#define DECOMPRESS_ADVANCED_HPP


#include "construct.hpp"
#include "sdsl/io.hpp"
#include <map>
#include <sstream>
#include <iostream>
#include "common.hpp"

#define SPECIALCHAR 1

void decompress_advanced(std::string input, std::string output, std::string comp)
{
    using namespace std;
    using namespace sdsl;
    // Decompress outer file	
    stringstream s;
	s << "tar -xf " << input;
        
    auto r = system(s.str().c_str());
	if (r == -1) cerr << "Error while untar all" << endl;
	
	std::string folder = ".";
	// Decompress files
	if (!std::strncmp(comp.c_str(), "ppmd", 4))
	{	
	    std::string command = "ppmd d -d -s " + folder + "/mapping.tar.pmd";
	    r = system(command.c_str());
	    if (r != 0) cerr << "Err while unppmd mapping" << endl;

	    command = "ppmd d -d -s " + folder + "/F.sdsl.pmd";
	    r = system(command.c_str());
	    if (r != 0) cerr << "Error while unppmd F" << endl;
	
	    command = "ppmd d -d -s " + folder + "/S.sdsl.pmd";
	    r = system(command.c_str());
	    if (r != 0) cerr << "Error while unppmd S" << endl;
    }
	else if (!std::strncmp(comp.c_str(), "zpaq", 4))
	{
	    std::string command = "~/zpaq/zpaq x " + folder + "/mapping.tar.zpaq mapping.tar -m5 -t1" ;
	    r = system(command.c_str());
	    if (r != 0) cerr << "Err while unzpaq mapping" << endl;

	    command = "~/zpaq/zpaq x " + folder + "/F.sdsl.zpaq F.sdsl -m5 -t1";
	    r = system(command.c_str());
	    if (r != 0) cerr << "Error while unzpaq F" << endl;
	
	    command = "~/zpaq/zpaq x " + folder + "/S.sdsl.zpaq S.sdsl -m5 -t1";
	    r = system(command.c_str());
	    if (r != 0) cerr << "Error while unzpaq S" << endl;
	}
	else
	{
	    std::string command = "xz -d -9 " + folder + "/mapping.tar.xz";
	    r = system(command.c_str());
	    if (r != 0) cerr << "Err while unxz mapping" << endl;

	    command = "xz -d -9 " + folder + "/F.sdsl.xz";
	    r = system(command.c_str());
	    if (r != 0) cerr << "Error while unxz F" << endl;
	
	    command = "xz -d -9 " + folder + "/S.sdsl.xz";
	    r = system(command.c_str());
	    if (r != 0) cerr << "Error while unxz S" << endl;
    }
    std::string command = "tar -xf " + folder + "/mapping.tar && rm " + folder + "/mapping.tar";
	r = system(command.c_str());
	if (r != 0) cerr << "Error while untar mapping" << endl;

	int_vector<8> S;
	load_from_file(S, folder + "/S.sdsl");
	
	int_vector<> F;
	load_from_file(F, folder + "/F.sdsl"); 
	
	// Load mapping start
	int_vector<> abs_start;	
	load_from_file(abs_start, folder + "/mapping_start_abs.sdsl");
	bit_vector bv_start;
	load_from_file(bv_start, folder + "/mapping_start_bv.sdsl");
	int_vector<> rel_start;	
	load_from_file(rel_start, folder + "/mapping_start_rel.sdsl");
	tuple<uint64_t, uint64_t, uint64_t> idx_start(1, 0, 0);
	
	get<1>(idx_start) += bv_start[0];	// Skip first value in abs ( eq. p ) if method == 1


	// Load mapping length
	int_vector<> long_len;
	load_from_file(long_len, folder + "/mapping_length_long.sdsl");
	int_vector<> short_len;
	load_from_file(short_len, folder + "/mapping_length_short.sdsl");
	bit_vector bv_len;
	load_from_file(bv_len, folder + "/mapping_length_bv.sdsl");
	
	
	tuple<uint64_t, uint64_t, uint64_t> idx_len(1, 0, 0);
	get<1>(idx_len) += bv_len[0];	// Skip first value in long ( eq. p ) if method == 1

    //cout << mapping_start_stream << endl;
    //cout << mapping_length_stream << endl;
    
    vector<uint64_t> mapping_start;
    vector<uint64_t> mapping_length;
    uint64_t prev_end = 0;
        
    //cout << mapping_start.size() << endl;
    
    vector<char> S_re;
    
    uint64_t f_id = 0;
    uint64_t curpos = 0;
    uint64_t curpos2 = 0;
    uint64_t uid;
    char c = 0;
    char specialchar = SPECIALCHAR;
    
    uint64_t size = S.size();
    for (uint64_t i = 0; i < size; i++) 
    {   
        c = S[i];
                
        if (i == 0)
        {
            specialchar = c;
            continue;
        }
        
        if (c == specialchar)
        {   // Special Char
            uid = F[f_id];
            f_id++;
                        
            /*
            uid = S[++i];
            uid = uid << 8;
            uid += S[++i];
            */
            
            if (uid == 0)       // Special char was in original string
            {
                S_re.push_back(c);
                curpos++;
                continue;
            }
            
            if (uid == 1)  // Special case aka repeat
            {
            	uint64_t t = get_next_mapping_start(idx_start, bv_start, abs_start, rel_start, prev_end);
                //uint64_t t = mapping_start_stream[next];
                uint64_t l = get_next_mapping_len(idx_len, bv_len, long_len, short_len);
                
                uint64_t cur = S_re.size();
                
                for (uint64_t i=0; i<l; i++)
                {
                    S_re.push_back(S_re[cur-t+i]);
                }
                
                prev_end = t + l;
                
                continue;
            }
            
            uint64_t s, end = 0;
            
            if (uid == 2)            
            //if (mapping_start.size() <= uid-3)       // new factor
            {
            	s = get_next_mapping_start(idx_start, bv_start, abs_start, rel_start, prev_end);
            	uint64_t len = get_next_mapping_len(idx_len, bv_len, long_len, short_len);
                mapping_start.push_back( s );
                mapping_length.push_back( len );
                                
                end = s + len;
                prev_end = end;
            }
            else
            {
		        // uid-3, because 0, 1, 2 are special cases
		        s = mapping_start[uid-3];
		        end = s + mapping_length[uid-3];
	     	} 
	     	
            for (; s < end; s++)
            {
                S_re.push_back(S_re[s]);
                curpos++;
            }
            
        }
        else 
        {
            S_re.push_back(c);
            curpos++;
        }
        curpos2++;
    }
    	
    ofstream f;
    f.open (output);
    f << std::string(S_re.begin(), S_re.end());
    f.close();
    
    // Remove tmp files
    r = system("rm S.sdsl F.sdsl mapping_start_abs.sdsl mapping_start_rel.sdsl mapping_start_bv.sdsl "
               "mapping_length_short.sdsl mapping_length_long.sdsl mapping_length_bv.sdsl");
	if (r == -1) cerr << "Error while rm tmp files" << endl;
    
    return;
}

#endif
